#include <asm/io.h>
#include <asm/system.h>
#include <kernel.h>
#include <floppy.h>
#include <display.h>
#include <irq.h>
#include <dma.h>
#include <timer.h>
#include <string.h>

#define FLOPPY_DMA     2

struct floppy
{
	unsigned int sect,
				 head,
				 track,
				 max_sect,
				 gap;
};

void irq_floppy(void);

static int motor_on(void);
static void motor_off(int id);
static unsigned char get_byte(void);
static void send_byte(unsigned char byte);
static void dma_init(int mode, int count);

static struct floppy fl = { 18, 2, 80, 18*2*80, 0x1b };
static volatile int flags;
static char floppy_buff[SIZE_SECTOR*3];

void do_floppy(void)
{
	outb(0x20,0x20);
	flags = 1;
}

unsigned char get_byte(void)
{
	int repeat;

	for( repeat=0; repeat < 10000; repeat++ )
		if( (inb(0x3f4) & 0xc0) == 0xc0 )
			return inb(0x3f5);

	puts("fp: get_byte ERROR\n");
	return -1;
}

void send_byte(unsigned char byte)
{
	int repeat;

	for( repeat=0; repeat < 10000; repeat++ )
		if( (inb(0x3f4) & 0xc0) == 0x80 )
		{
			outb(byte,0x3f5);
			return;
		}

	puts("fp: send_byte ERROR\n");
}

void dma_init(int mode, int count)
{
	unsigned int eflags = get_flags();

	cli();
	{
		disable_dma(FLOPPY_DMA);
		set_dma_mode(FLOPPY_DMA,mode);
		set_dma_addr(FLOPPY_DMA,(unsigned int)floppy_buff);
		set_dma_count(FLOPPY_DMA,count);
		enable_dma(FLOPPY_DMA);
	}
	set_flags(eflags);
}

#define ID_MOTOR_TIMER   1
#define ID_MOTOR_WAIT    2

static volatile char fwait;
/*    */
static int cylinder;
static int mt;

void motor_off(int id)
{
	outb(0,0x3f2);
	cylinder = -1;
	mt = 0;
}

void motor_wait(int id)
{
	fwait = 0;
}

int motor_on(void)
{
	if( !reset_timer(ID_MOTOR_TIMER,36) )
		return 0;

	outb( 0x1c,0x3f2);
	fwait = 1;
	set_timer(ID_MOTOR_WAIT,10,motor_wait);
	while( fwait );
/*
	if( inb(0x3f7) & 0x80 )
	{
		outb(0,0x3f2);
		return -1;
	}
*/
	set_timer(ID_MOTOR_TIMER,54,motor_off);
	mt = 1;
	return 0;
}

/*    floppy */
static void send_cmd(const unsigned char *cmd, int size)
{
	int i;

	for( i=0; i < size; i++ )
		send_byte(cmd[i]);
}

/*     */
static void get_rsl(unsigned char *rsl, int size)
{
	int i;

	for( i=0; i < size; i++ )
		rsl[i] = get_byte();
}

static int disk_not;

static int recl_seek(int fl, int ncn)
{
	unsigned char cmd[3];

	motor_on();

	flags = 0;
	if( fl )
	{
		/* SEEK */
		cmd[0] = 0x0f;
		cmd[1] = 0;
		cmd[2] = ncn;
		send_cmd(cmd,3);
	}
	else
	{
		/* RECALIBRATE */
		cmd[0] = 0x07;
		cmd[1] = 0;
		send_cmd(cmd,2);
	}

	while( !flags );

	/* SENSE INTERRUPT STATUS */
	cmd[0] = 0x08;
	send_cmd(cmd,1);
	get_rsl(cmd,2);

	if( inb(0x3f7) & 0x80 )
	{
		puts("fp: disk not find.\n");
		disk_not = 1;
		return -1;
	}
	else disk_not = 0;

	/*   */
	if( (cmd[0] & 0xe0) != 0x20 )
	{
		printk("fp: recl_seek ERROR\n"
				"st0 = 0x%x; pcn = 0x%x\n",
				cmd[0],cmd[1]);
		return -1;
	}
	return 0;
}

#define FDC_8272     0x80   /*  8272 */
#define FDC_82077    0x90   /*  82077 */

void floppy_init(void)
{
	unsigned char ver;

	set_irq(6,irq_floppy);
	cylinder = -1;

	/*   */
	outb_p(0x0c,0x3f2);
	/*    */
	send_byte(0x10);
	ver = get_byte();
	/*   */
	outb_p(0,0x3f2);
	/*   */
	if (ver == FDC_8272)
		puts("floppy - 8272\n");
	else if (ver == FDC_82077)
		puts("floppy - 82077\n");
	else
		puts("floppy init\n");
}

int read_sect(int sect, char *buff)
{
	char cmd[9];
	int c,h,s;
	int cnt1=0,cnt2=0;

	/*    */
	if( sect >= fl.max_sect )
	{
		printk("floppy: param ERROR, sect = %d\n",sect);
		return -1;
	}

	/*   */
	c = sect / (fl.head * fl.sect);
	/*   */
	h = (sect - c * fl.head * fl.sect) / fl.sect;
	/*     */
	s = sect - (c * fl.head + h) * fl.sect + 1;

	motor_on(); 
	if( disk_not )
	{
		recl_seek(1,1);
		cylinder = -1;
	}
	if( c != cylinder )
	{
		cylinder = c;
		if( recl_seek(0,0) )
		{
			puts("ERROR recl\n");
			return -1;
		}
		if( recl_seek(1,c) )
		{
			puts("ERROR seek\n");
			return -1;
		}
	}

LOOP_1:
	dma_init(DMA_MODE_READ,511);

	flags = 0;
	cmd[0] = 0x46;
	cmd[1] = h<<2;
	cmd[2] = c;
	cmd[3] = h;
	cmd[4] = s;
	cmd[5] = 2;
	cmd[6] = 18;
	cmd[7] = fl.gap;
	cmd[8] = 0xff;
	send_cmd(cmd,9);

	while( !flags );

	get_rsl(cmd,7);

	if( (cmd[0] & 0xc0) != 0 )
	{
		if( ++cnt1 < 3 )
		{
			puts("fl: read GOTO\n");
			goto LOOP_1;
		}
		cnt1 = 0;

		recl_seek(0,0);

		if( ++cnt2 > 3 )
		{
			puts("fl: read ERROR\n");
			return -1;
		}

		puts("fl: seek GOTO\n");
		cylinder = c;
		recl_seek(0,0);
		recl_seek(1,c);
		goto LOOP_1;
	}

	/*    */
	memcpy(buff,floppy_buff,512);

	return 0;
}

void test_sect(unsigned char *buff)
{
	recl_seek(0,0);
	recl_seek(1,0);
}

int write_sect(int sect, char *buff)
{
	char cmd[9];
	int c,h,s;
	int cnt1=0,cnt2=0;

	if (sect >= fl.max_sect) {
		printk("fl: param ERROR, sect = %d\n",sect);
		return -1;
	}

	/*   */
	c = sect / (fl.head * fl.sect);
	/*   */
	h = (sect - c * fl.head * fl.sect) / fl.sect;
	/*     */
	s = sect - (c * fl.head + h) * fl.sect + 1;

	motor_on();
	if (disk_not) {
		recl_seek(1,1);
		cylinder = -1;
	}
	if (c != cylinder) {
		cylinder = c;
		if (recl_seek(0,0)) {
			puts("ERROR recl\n");
			return -1;
		}
		if (recl_seek(1,c)) {
			puts("ERROR seek\n");
			return -1;
		}
	}

LOOP_1:
	memcpy(floppy_buff,buff,512);
	dma_init(DMA_MODE_WRITE,511);

	flags = 0;
	cmd[0] = 0x45;
	cmd[1] = (h << 2);
	cmd[2] = c;
	cmd[3] = h;
	cmd[4] = s;
	cmd[5] = 2;
	cmd[6] = fl.sect;
	cmd[7] = fl.gap;
	cmd[8] = 0xff;
	send_cmd(cmd,9);

	while (!flags);

	get_rsl(cmd,7);

	if ((cmd[0] & 0xc0) != 0) {
		if (++cnt1 < 3) {
			puts("fl: write GOTO\n");
			goto LOOP_1;
		}
		cnt1 = 0;

		if (++cnt2 > 3) {
			puts("fl: write ERROR\n");
			return -1;
		}

		puts("fl: seek GOTO\n");
		cylinder = c;
		recl_seek(0,0);
		recl_seek(1,c);
		goto LOOP_1;
	}

	return 0;
}

int format_sect(int cyl)
{
	char h;
	char cmd[7];
	char f = 0;
	int m;

	if (cyl >= fl.track) {
		printk("fl: param ERROR, cyl = %d\n",cyl);
		return -1;
	}

	m = mt;
	motor_on();
	if (disk_not) {
		recl_seek(1,1);
		cylinder = -1;
	}
	if (!m) {
		if (recl_seek(0,0)) {
			puts("ERROR recl\n");
			return -1;
		}
	}
	if (recl_seek(1,cyl)) {
		puts("ERROR seek\n");
		return -1;
	}

	h = 0;

	printk("cyl = %d\n",cyl);

LOOP_1:
	{
		int count;
		unsigned char *p = floppy_buff;

		for (count = 0; count < 18; count++) {
			*p++ = cyl;
			*p++ = h;
			*p++ = 1 + count;
			*p++ = 2;
		}
	}
	dma_init(DMA_MODE_WRITE,18*4);

	flags = 0;
	cmd[0] = 0x4d;
	cmd[1] = (h << 2);
	cmd[2] = 2;
	cmd[3] = 18;
	cmd[4] = 0x6c;
	cmd[5] = 0;
	send_cmd(cmd,6);

	while (!flags);

	get_rsl(cmd,7);
	printk("%x %x %x\n",cmd[0],cmd[1],cmd[2]);
	if (!f) {
		f = 1;
		h = 1;
		goto LOOP_1;
	}

	cylinder = -1;
	return 0;
}
